
from typing import Callable
import numpy as np
import pandas as pd

def resample(x: np.ndarray, prop: float)-> np.ndarray:
    """
    resample
    ========

    parameters:
        x    (numpy.ndarray[bool])
        prop (float)

    returns:
        numpy.ndarray[int]

    Returns indices that can be used to subset a vector, to yield a new vector
    with a given proportion of True values.
    """
    selection_mask = np.full(x.shape[0], True)

    ratio = prop / x.mean()

    if ratio > 1:
        ratio = 1-(ratio - 1)
        to_subset = ~x
    else:
        to_subset = x

    print(selection_mask.sum())
    selection_mask[to_subset] = np.random.choice([True, False], to_subset.sum(), p = [ratio, 1-ratio])
    print(selection_mask.sum())

    return np.linspace(0,x.shape[0]-1, x.shape[0], dtype = int)[selection_mask]

def resample_df(
        dataframe: pd.DataFrame,
        prop: float,
        column: str,
        key: Callable[[np.ndarray], np.ndarray] = lambda x: x > 0)-> pd.DataFrame:
    """
    resample_df
    ===========

    parameters:
        dataframe (pandas.DataFrame)
        prop      (float)
        column    (str)
        key       (Callable[[np.ndarray[Any]], np.ndarray[bool]])

    Resamples a dataframe to have a given proportion of values that return true
    after a key function is called The default key function is x > 0.
    """

    return dataframe.loc[resample(key(dataframe[column]), prop)]




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['views_runs']

package_data = \
{'': ['*']}

install_requires = \
['sklearn>=0.0,<0.1',
 'stepshift>=2.0.0,<2.1.0',
 'views-partitioning>=3.0.0,<4.0.0',
 'viewser>=5.5.0,<6.0.0']

setup_kwargs = {
    'name': 'views-runs',
    'version': '1.1.0',
    'description': '',
    'long_description': None,
    'author': 'peder2911',
    'author_email': 'pglandsverk@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



"""
views_runs.utilities
====================

Some utility functions ported from ViEWS 2, authored by Frederick Hoyles.

    resample:             Alter the distribution of binary values in a dataframe
    balance_panel_last_t: Ensure balanced panel form for dataframe
    assign_into_df:       Assign values from one dataframe into another
    merge_dfs:            Merge a list of unit-time indexed dataframes into one

"""
from typing import List
import functools

import numpy as np
import pandas as pd

def resample(
    df: pd.DataFrame,
    cols: List[str],
    pst_positives: float,
    pst_negatives: float,
    threshold=0,
):
    """
    resample
    ========

    parameters:
        df            (pandas.DataFrame)
        cols          (List[str])
        pst_positives (float)
        pst_negatives (float)
        threshold     (int)

    returns:
        pandas.DataFrame

    This function lets you resample a dataframe based on the desired retained
    percentage of positives and negatives in the column(s) specified in cols.

    Threshold is an integer that can be used to adjust what is defined as a
    "positive". With a threshold of 10, for example, the percentages affect
    values above and below 10, respectively.
    """

    # If both shares are 1 just return the unaltered df
    if pst_positives == 1 and pst_negatives == 1:
        return df

    # Negatives are rows where all cols are close to zero
    mask_negatives = np.isclose(df[cols], threshold).max(axis=1)
    # Positives are all the others
    mask_positives = ~mask_negatives

    df_positives = df.loc[mask_positives]
    df_negatives = df.loc[mask_negatives]

    len_positives = len(df_positives)
    len_negatives = len(df_negatives)

    n_positives_wanted = int(pst_positives * len_positives)
    n_negatives_wanted = int(pst_negatives * len_negatives)

    replacement_pos = pst_positives > 1
    replacement_neg = pst_negatives > 1
    df = pd.concat(
        [
            df_positives.sample(n=n_positives_wanted, replace=replacement_pos),
            df_negatives.sample(n=n_negatives_wanted, replace=replacement_neg),
        ]
    )
    return df

def balance_panel_last_t(df: pd.DataFrame) -> pd.DataFrame:
    """
    balance_panel_last_t
    ====================

    parameters:
        df (pandas.DataFrame): TIME-UNIT indexed dataframe

    returns:
        pandas:DataFrame

    "Balances" a time-unit indexed dataframe, making sure that all UNITs
    present at the last T are present for all T.
    """

    return df.reindex(
        pd.MultiIndex.from_product(
            [
                df.index.levels[0].unique(),
                df.loc[df.index.levels[0].max()].index.unique(),
            ],
            names=df.index.names,
        )
    ).sort_index()


def assign_into_df(df_to: pd.DataFrame, df_from: pd.DataFrame) -> pd.DataFrame:
    """
    assign_into_df
    ==============

    parameters:
        df_to   (pandas.DataFrame)
        df_from (pandas.DataFrame)

    returns:
        pandas.DataFrame

    Only assigns non-missing values from df_from, meaning the same column can
    be inserted multiple times and values be retained if the row coverage is
    different between calls.  So a df_a with col_a covering months 100-110 and
    df_b with col_a covering months 111-120 could be assigned into a single df
    which would get values of col_a for months 100 - 120.
    """

    for col in df_from:
        # Get a Series of the col for all rows
        s = df_from.loc[:, col]

        # Get the "is not null" boolean series to use as mask, ~ is NOT
        mask = ~s.isnull()

        # Get the index from that mask,
        # ix is now index labels of rows with (not missing) data
        ix = s.loc[mask].index
        try:
            df_to.loc[ix, col] = s.loc[ix]
        # Handle KeyError when assigning into a df without all times or locations
        except KeyError:
            try:
                # Only assign ix positions that exist in df_to
                ix = pd.MultiIndex.from_tuples(
                    [tup for tup in ix if tup in df_to.index]
                )
                df_to.loc[ix, col] = s.loc[ix]
            # If it still doesn't work we've got no ix in common.
            except TypeError:
                raise TypeError("No index positions in common.")

    return df_to

def merge_dfs(dfs: List[pd.DataFrame], how="inner")-> pd.DataFrame:
    """
    merge_dfs
    =========

    parameters:
        dfs (List[pandas.DataFrame])

    returns:
        pandas.DataFrame

    Merge a list of data frames using the dataframe indices.
    """

    return functools.reduce(
        lambda left, right: pd.merge(
            left, right, left_index=True, right_index=True, how=how
        ),
        dfs,
    )



import functools
import numpy as np
import pandas as pd

class ValidationError(Exception):
    pass

def dataframe_is_right_format(dataframe: pd.DataFrame):
    try:
        assert len(dataframe.index.levels) == 2
    except AssertionError:
        raise ValidationError("Dataframe must have a two-level index")

    try:
        assert set(dataframe.dtypes) == {np.dtype(float)}
    except AssertionError:
        raise ValidationError("The dataframe must contain only np.float64 floats")

def views_validate(fn):
    @functools.wraps(fn)
    def inner(*args,**kwargs):
        dataframe_is_right_format(args[-1])
        return fn(*args,**kwargs)
    return inner




from typing import Any
from viewser.storage import model_object

_model_object_store = model_object.ModelObjectStorage()
_metadata_store     = model_object.ModelMetadataStorage()

def store(name: str, model: Any, metadata: Any):
    _model_object_store.write(name, model)
    _metadata_store.write(name, metadata)

def retrieve(name: str):
    return _model_object_store.read(name)

def list():
    return _model_object_store.list()

def fetch_metadata(name: str):
    return _metadata_store.read(name)




from stepshift.views import StepshiftedModels
from views_partitioning import DataPartitioner
import pandas as pd

from . import validation

class ViewsRun():
    """
    ViewsRun

    This class combines modelling and partitioning, delivering predictions in a
    familiar format and concise API.

    Takes two arguments, a views_partitioning.DataPartitioner instance, and a
    stepshift.StepshiftedModels instance. These instances determine how data is
    subset, and how modelling happens.
    """

    def __init__(self, partitioner: DataPartitioner, models: StepshiftedModels):
        self._models = models
        self._shifted_partitioner = partitioner.shift_left(self._models._steps_extent)
        self._partitioner = partitioner

    @validation.views_validate
    def fit(self, partition_name: str, timespan_name: str, data: pd.DataFrame)-> None:
        """
        Fit the model to the named partition. partition_name must be present in
        the DataPartitioner.
        """

        self._models.fit(self._shifted_partitioner(partition_name,timespan_name,data))

    @validation.views_validate
    def predict(self, partition_name: str, timespan_name: str, data: pd.DataFrame)-> pd.DataFrame:
        """
        Predict with data from the named partition. partition_name must be
        present in the DataPartitioner.
        """

        predictions = self._models.predict(
                self._shifted_partitioner(partition_name, timespan_name, data),
                combine = False)
        predictions.index.names = data.index.names
        data = data.merge(predictions, how = "left", left_index = True, right_index = True)
        return self._partitioner(partition_name,timespan_name,data)

    @property
    def models(self):
        return self._models.models
